"""This module is horrific so far, I had to speed up the writing of this module to gain the
absolute minimal confidence that our WASM support isn't completely broken.
Each test is split into two functions to retrieve run_in_pyodide output(...) especially
the coverage aspect!
"""

from __future__ import annotations

import uuid

import pytest

# Handle the import safely. If pytest-pyodide is not installed (e.g. checks in IDE),
# we define a dummy decorator and mock fixtures to avoid NameError/fixture errors.
try:
    from pytest_pyodide import run_in_pyodide
except ImportError:

    def run_in_pyodide(*args, **kwargs):
        def decorator(func):
            return pytest.mark.skip(reason="pytest-pyodide not installed")(func)

        return decorator

    @pytest.fixture
    def selenium():
        pytest.skip("pytest-pyodide not installed")

    @pytest.fixture
    def selenium_jspi():
        pytest.skip("pytest-pyodide not installed")


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_basic_get(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("https://httpbingo.org/get")
            assert response.status_code == 200
            assert "headers" in response.json()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_basic_get(selenium_jspi):
    """Test basic GET request works."""
    data = _inner_test_sync_basic_get(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_basic_post(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.post("https://httpbingo.org/post", json={"test": "data"})
            assert response.status_code == 200
            data = response.json()
            assert data["json"] == {"test": "data"}
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_basic_post(selenium_jspi):
    """Test basic POST request works."""
    data = _inner_test_sync_basic_post(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_streaming_read(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("https://httpbingo.org/stream/5", stream=True)
            assert response.status_code == 200

            chunks = []
            for chunk in response.iter_lines():
                if chunk:
                    chunks.append(chunk)

            assert len(chunks) == 5
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_streaming_read(selenium_jspi):
    """Test streaming response works."""
    data = _inner_test_sync_streaming_read(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_retry_works(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session(retries=3) as s:
            response = s.get("https://httpbingo.org/get")
            assert response.status_code == 200
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_retry_works(selenium_jspi):
    """Test that retry mechanism works."""
    data = _inner_test_sync_retry_works(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_basic_get(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("https://httpbingo.org/get")
            assert response.status_code == 200
            assert "headers" in response.json()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_basic_get(selenium):
    """Test basic async GET request works."""
    data = _inner_test_async_basic_get(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_basic_post(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.post("https://httpbingo.org/post", json={"test": "data"})
            assert response.status_code == 200
            data = response.json()
            assert data["json"] == {"test": "data"}
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_basic_post(selenium):
    """Test basic async POST request works."""
    data = _inner_test_async_basic_post(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_streaming_read(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("https://httpbingo.org/stream/5", stream=True)
            assert response.status_code == 200

            chunks = []
            async for chunk in response.iter_lines():
                if chunk:
                    chunks.append(chunk)

            assert len(chunks) == 5
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_streaming_read(selenium):
    """Test async streaming response works."""
    data = _inner_test_async_streaming_read(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_retry_works(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession(retries=3) as s:
            response = await s.get("https://httpbingo.org/get")
            assert response.status_code == 200
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_retry_works(selenium):
    """Test that async retry mechanism works."""
    data = _inner_test_async_retry_works(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_websocket(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import sys

    if "emscripten" not in sys.platform:
        return b""

    try:
        from js import WebSocket  # noqa: F401
    except ImportError:
        import pytest

        pytest.skip("websocket support unavailable on platform")

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("wss://echo.websocket.org")
            assert response.status_code == 101

            ext = response.extension
            assert ext is not None
            assert ext.closed is False

            # Consume potential welcome/greeting message from the server
            welcome = ext.next_payload()
            assert isinstance(welcome, str)

            ext.send_payload("hello from niquests")
            msg = ext.next_payload()
            assert isinstance(msg, str)
            assert msg == "hello from niquests"

            ext.close()
            assert ext.closed is True
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_websocket(selenium_jspi):
    """Test sync WebSocket via browser native API + JSPI."""
    data = _inner_test_sync_websocket(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_websocket_binary(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    try:
        from js import WebSocket  # noqa: F401
    except ImportError:
        import pytest

        pytest.skip("websocket support unavailable on platform")

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("wss://echo.websocket.org")
            assert response.status_code == 101

            ext = response.extension

            # Consume potential welcome/greeting message from the server
            ext.next_payload()

            ext.send_payload(b"\x00\x01\x02\x03")
            msg = ext.next_payload()
            assert isinstance(msg, bytes)
            assert msg == b"\x00\x01\x02\x03"

            ext.close()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_websocket_binary(selenium_jspi):
    """Test sync WebSocket binary message handling."""
    data = _inner_test_sync_websocket_binary(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_websocket_close_from_server(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    try:
        from js import WebSocket  # noqa: F401
    except ImportError:
        import pytest

        pytest.skip("websocket support unavailable on platform")

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("wss://echo.websocket.org")
            ext = response.extension
            assert ext.closed is False

            # Close from our side, then verify state
            ext.close()
            assert ext.closed is True
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_websocket_close_from_server(selenium_jspi):
    """Test that close works and state is updated."""
    data = _inner_test_sync_websocket_close_from_server(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_websocket(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    try:
        from js import WebSocket  # noqa: F401
    except ImportError:
        import pytest

        pytest.skip("websocket support unavailable on platform")

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("wss://echo.websocket.org")
            assert response.status_code == 101

            ext = response.extension
            assert ext is not None
            assert ext.closed is False

            # Consume potential welcome/greeting message from the server
            welcome = await ext.next_payload()
            assert isinstance(welcome, str)

            await ext.send_payload("hello from niquests async")
            msg = await ext.next_payload()
            assert isinstance(msg, str)
            assert msg == "hello from niquests async"

            await ext.close()
            assert ext.closed is True
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_websocket(selenium):
    """Test async WebSocket via browser native API."""
    data = _inner_test_async_websocket(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_websocket_binary(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    try:
        from js import WebSocket  # noqa: F401
    except ImportError:
        import pytest

        pytest.skip("websocket support unavailable on platform")

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("wss://echo.websocket.org")
            assert response.status_code == 101

            ext = response.extension

            # Consume potential welcome/greeting message from the server
            await ext.next_payload()

            await ext.send_payload(b"\x00\x01\x02\x03")
            msg = await ext.next_payload()
            assert isinstance(msg, bytes)
            assert msg == b"\x00\x01\x02\x03"

            await ext.close()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_websocket_binary(selenium):
    """Test async WebSocket binary message handling."""
    data = _inner_test_async_websocket_binary(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_sse(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session
        from niquests.packages.urllib3.contrib.webextensions.sse import ServerSentEvent

        with Session() as s:
            response = s.get("sse://httpbingo.org/sse?count=3&duration=1")
            assert response.status_code == 200

            ext = response.extension
            assert ext is not None
            assert ext.closed is False

            events = []
            while True:
                event = ext.next_payload()
                if event is None:
                    break
                assert isinstance(event, ServerSentEvent)
                events.append(event)

            assert len(events) >= 1
            assert ext.closed is True
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_sse(selenium_jspi):
    """Test sync SSE via pyfetch streaming + manual parsing."""
    data = _inner_test_sync_sse(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_sse(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession
        from niquests.packages.urllib3.contrib.webextensions.sse import ServerSentEvent

        async with AsyncSession() as s:
            response = await s.get("sse://httpbingo.org/sse?count=3&duration=1")
            assert response.status_code == 200

            ext = response.extension
            assert ext is not None
            assert ext.closed is False

            events = []
            while True:
                event = await ext.next_payload()
                if event is None:
                    break
                assert isinstance(event, ServerSentEvent)
                events.append(event)

            assert len(events) >= 1
            assert ext.closed is True
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_sse(selenium):
    """Test async SSE via pyfetch streaming + manual parsing."""
    data = _inner_test_async_sse(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_sse_send_raises(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import pytest

        from niquests import Session

        with Session() as s:
            response = s.get("sse://httpbingo.org/sse?count=1&duration=1")
            ext = response.extension

            with pytest.raises(NotImplementedError):
                ext.send_payload("test")

            ext.close()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_sse_send_raises(selenium_jspi):
    """Test that SSE send_payload raises NotImplementedError."""
    data = _inner_test_sync_sse_send_raises(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_timeout(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import pytest

        from niquests import Session, TimeoutConfiguration
        from niquests.exceptions import ConnectTimeout
        from niquests.packages.urllib3.exceptions import MaxRetryError

        with Session(retries=False) as s:
            with pytest.raises(ConnectTimeout):
                s.get("https://httpbingo.org/delay/5", timeout=0.5)

        with Session(retries=0) as s:
            with pytest.raises(MaxRetryError):
                s.get("https://httpbingo.org/delay/5", timeout=0.5)

        with Session(retries=0) as s:
            with pytest.raises(MaxRetryError):
                s.get("https://httpbingo.org/delay/5", timeout=TimeoutConfiguration(0.5))
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_timeout(selenium_jspi):
    """Test that timeout raises ConnectTimeout on slow responses."""
    data = _inner_test_sync_timeout(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_timeout(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import pytest

        from niquests import AsyncSession, TimeoutConfiguration
        from niquests.exceptions import ConnectTimeout
        from niquests.packages.urllib3.exceptions import MaxRetryError

        async with AsyncSession(retries=False) as s:
            with pytest.raises(ConnectTimeout):
                await s.get("https://httpbingo.org/delay/5", timeout=0.5)

        async with AsyncSession(retries=0) as s:
            with pytest.raises(MaxRetryError):
                await s.get("https://httpbingo.org/delay/5", timeout=0.5)

        async with AsyncSession(retries=0) as s:
            with pytest.raises(MaxRetryError):
                await s.get("https://httpbingo.org/delay/5", timeout=TimeoutConfiguration(0.5))
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_timeout(selenium):
    """Test that async timeout raises ConnectTimeout on slow responses."""
    data = _inner_test_async_timeout(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_custom_request_headers(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get(
                "https://httpbingo.org/headers",
                headers={"X-Custom-Test": "niquests-pyodide", "Accept": "application/json"},
            )
            assert response.status_code == 200

            data = response.json()
            # httpbingo returns the received headers
            headers = data.get("headers", {})
            assert "X-Custom-Test" in headers
            assert headers["X-Custom-Test"] == ["niquests-pyodide"]

            # Verify response headers are accessible
            assert "content-type" in response.headers
            assert "application/json" in response.headers["content-type"]
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_custom_request_headers(selenium_jspi):
    """Test that custom request headers are sent and response headers are parsed."""
    data = _inner_test_sync_custom_request_headers(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_custom_request_headers(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get(
                "https://httpbingo.org/headers",
                headers={"X-Custom-Test": "niquests-pyodide", "Accept": "application/json"},
            )
            assert response.status_code == 200

            data = response.json()
            headers = data.get("headers", {})
            assert "X-Custom-Test" in headers
            assert headers["X-Custom-Test"] == ["niquests-pyodide"]

            assert "content-type" in response.headers
            assert "application/json" in response.headers["content-type"]
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_custom_request_headers(selenium):
    """Test that async custom request headers are sent and response headers are parsed."""
    data = _inner_test_async_custom_request_headers(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_forbidden_header_silently_dropped(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get(
                "https://httpbingo.org/headers",
                headers={"Host": "evil.example.com", "Connection": "close", "X-Legit": "present"},
            )
            assert response.status_code == 200

            data = response.json()
            headers = data.get("headers", {})

            # The forbidden headers must NOT reach the server
            assert headers.get("Host", [None])[0] != "evil.example.com"
            assert headers.get("Connection", [""])[0] != "close"

            # Non-forbidden headers must still be sent
            assert "X-Legit" in headers
            assert headers["X-Legit"] == ["present"]
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_forbidden_header_silently_dropped(selenium_jspi):
    """Test that forbidden headers (e.g. Host) are silently stripped and do not cause errors."""
    data = _inner_test_sync_forbidden_header_silently_dropped(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_forbidden_header_silently_dropped(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get(
                "https://httpbingo.org/headers",
                headers={"Host": "evil.example.com", "Connection": "close", "X-Legit": "present"},
            )
            assert response.status_code == 200

            data = response.json()
            headers = data.get("headers", {})

            # The forbidden headers must NOT reach the server
            assert headers.get("Host", [None])[0] != "evil.example.com"
            assert headers.get("Connection", [""])[0] != "close"

            # Non-forbidden headers must still be sent
            assert "X-Legit" in headers
            assert headers["X-Legit"] == ["present"]
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_forbidden_header_silently_dropped(selenium):
    """Test that forbidden headers (e.g. Host) are silently stripped and do not cause errors (async)."""
    data = _inner_test_async_forbidden_header_silently_dropped(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_non_ok_response(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import pytest

        from niquests import Session
        from niquests.exceptions import HTTPError

        with Session() as s:
            response = s.get("https://httpbingo.org/status/418")
            assert response.status_code == 418
            assert response.ok is False

            with pytest.raises(HTTPError, match="418"):
                response.raise_for_status()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_non_ok_response(selenium_jspi):
    """Test non-OK responses and raise_for_status."""
    data = _inner_test_sync_non_ok_response(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_non_ok_response(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import pytest

        from niquests import AsyncSession
        from niquests.exceptions import HTTPError

        async with AsyncSession() as s:
            response = await s.get("https://httpbingo.org/status/418")
            assert response.status_code == 418
            assert response.ok is False

            with pytest.raises(HTTPError, match="418"):
                response.raise_for_status()
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_non_ok_response(selenium):
    """Test async non-OK responses and raise_for_status."""
    data = _inner_test_async_non_ok_response(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_binary_response(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("https://httpbingo.org/image/jpeg")
            assert response.status_code == 200
            assert "image/jpeg" in response.headers.get("content-type", "")

            # JPEG files start with FF D8
            assert isinstance(response.content, bytes)
            assert len(response.content) > 100
            assert response.content[:2] == b"\xff\xd8"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_binary_response(selenium_jspi):
    """Test binary response content (JPEG image)."""
    data = _inner_test_sync_binary_response(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_binary_response(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("https://httpbingo.org/image/jpeg")
            assert response.status_code == 200
            assert "image/jpeg" in response.headers.get("content-type", "")

            assert isinstance(response.content, bytes)
            assert len(response.content) > 100
            assert response.content[:2] == b"\xff\xd8"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_binary_response(selenium):
    """Test async binary response content (JPEG image)."""
    data = _inner_test_async_binary_response(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_redirect(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("https://httpbingo.org/redirect/3")
            assert response.status_code == 200
            # After 3 redirects, we should end up at /get
            assert response.url.endswith("/get") or "/get" in response.url
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_redirect(selenium_jspi):
    """Test that redirects are followed and final URL is correct."""
    data = _inner_test_sync_redirect(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_redirect(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("https://httpbingo.org/redirect/3")
            assert response.status_code == 200
            assert response.url.endswith("/get") or "/get" in response.url
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_redirect(selenium):
    """Test that async redirects are followed and final URL is correct."""
    data = _inner_test_async_redirect(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_string_body(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.post(
                "https://httpbingo.org/post",
                data="raw string body",
                headers={"Content-Type": "text/plain"},
            )
            assert response.status_code == 200
            data = response.json()
            assert data.get("data") == "raw string body"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_string_body(selenium_jspi):
    """Test POST with a raw string body."""
    data = _inner_test_sync_string_body(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_string_body(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.post(
                "https://httpbingo.org/post",
                data="raw string body",
                headers={"Content-Type": "text/plain"},
            )
            assert response.status_code == 200
            data = response.json()
            assert data.get("data") == "raw string body"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_string_body(selenium):
    """Test async POST with a raw string body."""
    data = _inner_test_async_string_body(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_iterable_body(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        def body_gen():
            yield b"chunk1"
            yield b"chunk2"
            yield b"chunk3"

        with Session() as s:
            response = s.post(
                "https://httpbingo.org/post",
                data=body_gen(),
                headers={"Content-Type": "application/octet-stream"},
            )
            assert response.status_code == 200
            data = response.json()
            assert "Y2h1bmsxY2h1bmsyY2h1bmsz" in data.get("data", "")
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_iterable_body(selenium_jspi):
    """Test POST with a generator/iterable body."""
    data = _inner_test_sync_iterable_body(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_iterable_body(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        def body_gen():
            yield b"chunk1"
            yield b"chunk2"
            yield b"chunk3"

        async with AsyncSession() as s:
            response = await s.post(
                "https://httpbingo.org/post",
                data=body_gen(),
                headers={"Content-Type": "application/octet-stream"},
            )
            assert response.status_code == 200
            data = response.json()
            assert "Y2h1bmsxY2h1bmsyY2h1bmsz" in data.get("data", "")
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_iterable_body(selenium):
    """Test async POST with a generator/iterable body."""
    data = _inner_test_async_iterable_body(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_iterable_str_body(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        def str_body_gen():
            yield "hello "
            yield "world"

        with Session() as s:
            response = s.post(
                "https://httpbingo.org/post",
                data=str_body_gen(),
                headers={"Content-Type": "text/plain"},
            )
            assert response.status_code == 200
            data = response.json()
            assert data.get("data") == "hello world"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_iterable_str_body(selenium_jspi):
    """Test POST with a generator yielding str chunks."""
    data = _inner_test_sync_iterable_str_body(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_iterable_str_body(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        def str_body_gen():
            yield "hello "
            yield "world"

        async with AsyncSession() as s:
            response = await s.post(
                "https://httpbingo.org/post",
                data=str_body_gen(),
                headers={"Content-Type": "text/plain"},
            )
            assert response.status_code == 200
            data = response.json()
            assert data.get("data") == "hello world"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_iterable_str_body(selenium):
    """Test async POST with a generator yielding str chunks."""
    data = _inner_test_async_iterable_str_body(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_async_iterable_body(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async def body_gen():
            yield b"chunk1"
            yield b"chunk2"
            yield b"chunk3"

        async with AsyncSession() as s:
            response = await s.post(
                "https://httpbingo.org/post",
                data=body_gen(),
                headers={"Content-Type": "application/octet-stream"},
            )
            assert response.status_code == 200
            data = response.json()
            assert "Y2h1bmsxY2h1bmsyY2h1bmsz" in data.get("data", "")
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_async_iterable_body(selenium):
    """Test async POST with an async generator body (covers __aiter__ path)."""
    data = _inner_test_async_async_iterable_body(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_async_iterable_str_body(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async def str_body_gen():
            yield "hello "
            yield "world"

        async with AsyncSession() as s:
            response = await s.post(
                "https://httpbingo.org/post",
                data=str_body_gen(),
                headers={"Content-Type": "text/plain"},
            )
            assert response.status_code == 200
            data = response.json()
            assert data.get("data") == "hello world"
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_async_iterable_str_body(selenium):
    """Test async POST with an async generator yielding str chunks."""
    data = _inner_test_async_async_iterable_str_body(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_sse_raw_mode(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get("sse://httpbingo.org/sse?count=2&duration=1")
            assert response.status_code == 200

            ext = response.extension
            assert ext is not None

            events = []
            while True:
                event = ext.next_payload(raw=True)
                if event is None:
                    break
                assert isinstance(event, str)
                assert event.endswith("\n\n")
                events.append(event)

            assert len(events) >= 1
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_sse_raw_mode(selenium_jspi):
    """Test sync SSE with raw=True returns raw event strings."""
    data = _inner_test_sync_sse_raw_mode(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_sse_raw_mode(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get("sse://httpbingo.org/sse?count=2&duration=1")
            assert response.status_code == 200

            ext = response.extension
            assert ext is not None

            events = []
            while True:
                event = await ext.next_payload(raw=True)
                if event is None:
                    break
                assert isinstance(event, str)
                assert event.endswith("\n\n")
                events.append(event)

            assert len(events) >= 1
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_sse_raw_mode(selenium):
    """Test async SSE with raw=True returns raw event strings."""
    data = _inner_test_async_sse_raw_mode(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_sync_sse_via_stream(selenium_jspi):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import Session

        with Session() as s:
            response = s.get(
                "https://httpbingo.org/sse?count=3&duration=1",
                stream=True,
            )
            assert response.status_code == 200

            chunks = []
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    chunks.append(chunk)

            assert len(chunks) >= 1
            raw = b"".join(chunks)
            # SSE events contain "data:" fields
            assert b"data:" in raw
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_sync_sse_via_stream(selenium_jspi):
    """Test consuming an SSE endpoint as a plain stream (no sse:// scheme)."""
    data = _inner_test_sync_sse_via_stream(selenium_jspi)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_sse_via_stream(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        from niquests import AsyncSession

        async with AsyncSession() as s:
            response = await s.get(
                "https://httpbingo.org/sse?count=3&duration=1",
                stream=True,
            )
            assert response.status_code == 200

            chunks = []
            async for chunk in await response.iter_content(chunk_size=1024):
                if chunk:
                    chunks.append(chunk)

            assert len(chunks) >= 1
            raw = b"".join(chunks)
            # SSE events contain "data:" fields
            assert b"data:" in raw
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_sse_via_stream(selenium):
    """Test consuming an SSE endpoint as a plain async stream (no sse:// scheme)."""
    data = _inner_test_async_sse_via_stream(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)


@run_in_pyodide(packages=["micropip"])
async def _inner_test_async_concurrent_gather(selenium):
    import micropip

    await micropip.install(["./niquests-0.0.dev0-py3-none-any.whl", "coverage"], deps=True)

    import coverage

    cov = coverage.Coverage(source=["niquests"])
    cov.start()
    try:
        import asyncio

        from niquests import AsyncSession

        async with AsyncSession() as s:

            async def do_get(req_id):
                resp = await s.get(f"https://httpbingo.org/get?req={req_id}")
                return req_id, resp

            results = await asyncio.gather(
                do_get("a1"),
                do_get("b2"),
                do_get("c3"),
                do_get("d4"),
                do_get("e5"),
            )

            assert len(results) == 5

            seen_ids = set()
            for req_id, resp in results:
                assert resp.status_code == 200
                data = resp.json()
                assert data["args"]["req"] == [req_id]
                seen_ids.add(req_id)

            assert seen_ids == {"a1", "b2", "c3", "d4", "e5"}
    finally:
        cov.stop()
        cov.save()
    with open(".coverage", "rb") as f:
        return f.read()


def test_async_concurrent_gather(selenium):
    """Test that 5 concurrent async requests via asyncio.gather all complete."""
    data = _inner_test_async_concurrent_gather(selenium)
    if data:
        with open(f".coverage.pyodide.{uuid.uuid4().hex}", "wb") as f:
            f.write(data)



from __future__ import annotations

import time
import typing
from datetime import timedelta

from pyodide.ffi import run_sync  # type: ignore[import]
from pyodide.http import pyfetch  # type: ignore[import]

from ..._constant import DEFAULT_RETRIES
from ...adapters import BaseAdapter
from ...exceptions import ConnectionError, ConnectTimeout
from ...models import PreparedRequest, Response
from ...packages.urllib3.exceptions import MaxRetryError
from ...packages.urllib3.response import BytesQueueBuffer
from ...packages.urllib3.response import HTTPResponse as BaseHTTPResponse
from ...packages.urllib3.util import Timeout as TimeoutSauce
from ...packages.urllib3.util.retry import Retry
from ...structures import CaseInsensitiveDict
from ...utils import get_encoding_from_headers

if typing.TYPE_CHECKING:
    from ...typing import ProxyType, RetryType, TLSClientCertType, TLSVerifyType


class _PyodideRawIO:
    """File-like wrapper around a Pyodide Fetch response with true streaming via JSPI.

    When constructed with a JS Response object, reads chunks incrementally from the
    JavaScript ReadableStream using ``run_sync(reader.read())`` per chunk.
    When constructed with preloaded content (non-streaming), serves from a memory buffer.
    """

    def __init__(
        self,
        js_response: typing.Any = None,
        preloaded_content: bytes | None = None,
    ) -> None:
        self._js_response = js_response
        self._buffer = BytesQueueBuffer()
        self._closed = False
        self._finished = preloaded_content is not None or js_response is None
        self._reader: typing.Any = None
        self.headers: dict[str, str] = {}
        self.extension: typing.Any = None

        if preloaded_content is not None:
            self._buffer.put(preloaded_content)

    def _ensure_reader(self) -> None:
        """Initialize the ReadableStream reader if not already done."""
        if self._reader is None and self._js_response is not None:
            try:
                body = self._js_response.body
                if body is not None:
                    self._reader = body.getReader()
            except Exception:
                pass

    def _get_next_chunk(self) -> bytes | None:
        """Read the next chunk from the JS ReadableStream, blocking via JSPI."""
        self._ensure_reader()

        if self._reader is None:
            return None

        try:
            result = run_sync(self._reader.read())

            if result.done:
                return None

            value = result.value
            if value is not None:
                return bytes(value.to_py())
            return None
        except Exception:
            return None

    def read(
        self,
        amt: int | None = None,
        decode_content: bool = True,
    ) -> bytes:
        if self._closed:
            return self._buffer.get(len(self._buffer)) if len(self._buffer) > 0 else b""

        if self._finished:
            if len(self._buffer) == 0:
                return b""
            if amt is None or amt < 0:
                return self._buffer.get(len(self._buffer))
            return self._buffer.get(min(amt, len(self._buffer)))

        if amt is None or amt < 0:
            # Read everything remaining
            while True:
                chunk = self._get_next_chunk()
                if chunk is None:
                    break
                self._buffer.put(chunk)
            self._finished = True
            return self._buffer.get(len(self._buffer)) if len(self._buffer) > 0 else b""

        # Read until we have enough bytes or stream ends
        while len(self._buffer) < amt and not self._finished:
            chunk = self._get_next_chunk()
            if chunk is None:
                self._finished = True
                break
            self._buffer.put(chunk)

        if len(self._buffer) == 0:
            return b""

        return self._buffer.get(min(amt, len(self._buffer)))

    def stream(self, amt: int, decode_content: bool = True) -> typing.Generator[bytes, None, None]:
        """Iterate over chunks of the response."""
        while True:
            chunk = self.read(amt)
            if not chunk:
                break
            yield chunk

    def close(self) -> None:
        self._closed = True
        if self._reader is not None:
            try:
                run_sync(self._reader.cancel())
            except Exception:
                pass
        self._reader = None
        self._js_response = None

    def __iter__(self) -> typing.Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        chunk = self.read(8192)
        if not chunk:
            raise StopIteration
        return chunk


class PyodideAdapter(BaseAdapter):
    """Synchronous adapter for making HTTP requests in Pyodide using JSPI + pyfetch."""

    def __init__(self, max_retries: RetryType = DEFAULT_RETRIES) -> None:
        super().__init__()

        if isinstance(max_retries, Retry):
            self.max_retries = max_retries
        else:
            self.max_retries = Retry.from_int(max_retries)

    def __repr__(self) -> str:
        return "<PyodideAdapter WASM/>"

    def send(
        self,
        request: PreparedRequest,
        stream: bool = False,
        timeout: int | float | tuple | TimeoutSauce | None = None,
        verify: TLSVerifyType = True,
        cert: TLSClientCertType | None = None,
        proxies: ProxyType | None = None,
        on_post_connection: typing.Callable[[typing.Any], None] | None = None,
        on_upload_body: typing.Callable[[int, int | None, bool, bool], None] | None = None,
        on_early_response: typing.Callable[[Response], None] | None = None,
        multiplexed: bool = False,
    ) -> Response:
        """Send a PreparedRequest using Pyodide's pyfetch (synchronous via JSPI)."""
        if isinstance(timeout, tuple):
            if len(timeout) == 3:
                timeout = timeout[2] or timeout[0]  # prefer total, fallback connect
            else:
                timeout = timeout[0]  # use connect
        elif isinstance(timeout, TimeoutSauce):
            timeout = timeout.total or timeout.connect_timeout

        retries = self.max_retries
        method = request.method or "GET"
        start = time.time()

        while True:
            try:
                response = self._do_send(request, stream, timeout)
            except Exception as err:
                retries = retries.increment(method, request.url, error=err)

                retries.sleep()
                continue

            base_response = BaseHTTPResponse(
                body=b"",
                headers=response.headers,
                status=response.status_code,
                request_method=request.method,
                request_url=request.url,
            )

            has_retry_after = bool(response.headers.get("Retry-After"))

            if retries.is_retry(method, response.status_code, has_retry_after):
                try:
                    retries = retries.increment(method, request.url, response=base_response)
                except MaxRetryError:
                    if retries.raise_on_status:
                        raise
                    return response

                retries.sleep(base_response)
                continue

            response.elapsed = timedelta(seconds=time.time() - start)
            return response

    def _do_send(
        self,
        request: PreparedRequest,
        stream: bool,
        timeout: int | float | None,
    ) -> Response:
        """Perform the actual request using pyfetch made synchronous via JSPI."""
        url = request.url or ""
        scheme = url.split("://")[0].lower() if "://" in url else ""

        # WebSocket: delegate to browser native WebSocket API
        if scheme in ("ws", "wss"):
            return self._do_send_ws(request, url)

        # SSE: delegate to pyfetch streaming + manual SSE parsing
        if scheme in ("sse", "psse"):
            return self._do_send_sse(request, url, scheme)

        # Prepare headers
        headers_dict: dict[str, str] = {}
        if request.headers:
            for key, value in request.headers.items():
                if key.lower() not in ("host", "content-length", "connection", "transfer-encoding"):
                    headers_dict[key] = value

        # Prepare body
        body = request.body
        if body is not None:
            if isinstance(body, str):
                body = body.encode("utf-8")
            elif isinstance(body, typing.Iterable) and not isinstance(body, (bytes, bytearray)):
                chunks: list[bytes] = []
                for chunk in body:
                    if isinstance(chunk, str):
                        chunks.append(chunk.encode("utf-8"))
                    elif isinstance(chunk, bytes):
                        chunks.append(chunk)
                body = b"".join(chunks)

        # Build fetch options
        fetch_options: dict[str, typing.Any] = {
            "method": request.method or "GET",
            "headers": headers_dict,
        }

        if body:
            fetch_options["body"] = body

        # Use AbortSignal.timeout() for timeout — the browser-native mechanism.
        # run_sync cannot interrupt a JS Promise.
        signal = None

        if timeout is not None:
            from js import AbortSignal  # type: ignore[import]

            signal = AbortSignal.timeout(int(timeout * 1000))

        try:
            js_response = run_sync(pyfetch(request.url, signal=signal, **fetch_options))
        except Exception as e:
            err_str = str(e).lower()
            if "abort" in err_str or "timeout" in err_str or "timed out" in err_str:
                raise ConnectTimeout(f"Connection to {request.url} timed out")
            raise ConnectionError(f"Failed to fetch {request.url}: {e}")

        # Parse response headers
        response_headers: dict[str, str] = {}
        try:
            if hasattr(js_response, "headers"):
                js_headers = js_response.headers
                if hasattr(js_headers, "items"):
                    for key, value in js_headers.items():
                        response_headers[key] = value
                elif hasattr(js_headers, "entries"):
                    for entry in js_headers.entries():
                        response_headers[entry[0]] = entry[1]
        except Exception:
            pass

        # Build response object
        response = Response()
        response.status_code = js_response.status
        response.headers = CaseInsensitiveDict(response_headers)
        response.request = request
        response.url = js_response.url or request.url
        response.encoding = get_encoding_from_headers(response_headers)

        try:
            response.reason = js_response.status_text or ""
        except Exception:
            response.reason = ""

        if stream:
            # Streaming: pass the underlying JS Response to the raw IO
            # so it can read chunks incrementally via run_sync(reader.read())
            raw_io = _PyodideRawIO(js_response=js_response.js_response)
            raw_io.headers = response_headers
            response.raw = raw_io  # type: ignore
            response._content = False  # type: ignore[assignment]
            response._content_consumed = False
        else:
            # Non-streaming: read full body upfront
            try:
                response_body: bytes = run_sync(js_response.bytes())
            except Exception:
                response_body = b""

            raw_io = _PyodideRawIO(preloaded_content=response_body)
            raw_io.headers = response_headers
            response.raw = raw_io  # type: ignore
            response._content = response_body

        return response

    def _do_send_ws(self, request: PreparedRequest, url: str) -> Response:
        """Handle WebSocket connections via browser native WebSocket API."""
        from ._ws import PyodideWebSocketExtension

        try:
            ext = PyodideWebSocketExtension(url)
        except Exception as e:
            raise ConnectionError(f"WebSocket connection to {url} failed: {e}")

        response = Response()
        response.status_code = 101
        response.headers = CaseInsensitiveDict({"upgrade": "websocket", "connection": "upgrade"})
        response.request = request
        response.url = url
        response.reason = "Switching Protocols"

        raw_io = _PyodideRawIO()
        raw_io.extension = ext
        response.raw = raw_io  # type: ignore
        response._content = b""

        return response

    def _do_send_sse(self, request: PreparedRequest, url: str, scheme: str) -> Response:
        """Handle SSE connections via pyfetch streaming + manual parsing."""
        from ._sse import PyodideSSEExtension

        http_url = url.replace("sse://", "https://", 1) if scheme == "sse" else url.replace("psse://", "http://", 1)

        # Pass through user-provided headers
        headers_dict: dict[str, str] = {}
        if request.headers:
            for key, value in request.headers.items():
                if key.lower() not in ("host", "content-length", "connection"):
                    headers_dict[key] = value

        try:
            ext = PyodideSSEExtension(http_url, headers=headers_dict)
        except Exception as e:
            raise ConnectionError(f"SSE connection to {url} failed: {e}")

        response = Response()
        response.status_code = 200
        response.headers = CaseInsensitiveDict({"content-type": "text/event-stream"})
        response.request = request
        response.url = url
        response.reason = "OK"

        raw_io = _PyodideRawIO()
        raw_io.extension = ext
        response.raw = raw_io  # type: ignore
        response._content = False  # type: ignore[assignment]
        response._content_consumed = False

        return response

    def close(self) -> None:
        """Clean up adapter resources."""
        pass


__all__ = ("PyodideAdapter",)



from __future__ import annotations

import importlib
import importlib.abc
import importlib.machinery
import importlib.util
import sys
import typing

from ._compat import HAS_LEGACY_URLLIB3

# just to enable smooth type-completion!
if typing.TYPE_CHECKING:
    import charset_normalizer as chardet
    import urllib3

    charset_normalizer = chardet

    import idna  # type: ignore[import-not-found]

# Mapping of aliased package prefixes:
#   "niquests.packages.<alias>." to "<real-package>."
# Populated by the loop below, consumed by the import hook.
_ALIAS_TO_REAL: dict[str, str] = {}

# This code exists for backwards compatibility reasons.
# I don't like it either. Just look the other way. :)
for package in (
    "urllib3",
    "charset_normalizer",
    "idna",
    "chardet",
):
    to_be_imported: str = package

    if package == "chardet":
        to_be_imported = "charset_normalizer"
    elif package == "urllib3" and HAS_LEGACY_URLLIB3:
        to_be_imported = "urllib3_future"

    try:
        locals()[package] = __import__(to_be_imported)
    except ImportError:
        continue  # idna could be missing. not required!

    # Determine the alias prefix (what niquests code imports)
    # and the real prefix (the actual installed package).
    if package == "chardet":
        alias_prefix = "niquests.packages.chardet."
        real_prefix = "charset_normalizer."
        alias_root = "niquests.packages.chardet"
        real_root = "charset_normalizer"
    elif package == "urllib3" and HAS_LEGACY_URLLIB3:
        alias_prefix = "niquests.packages.urllib3."
        real_prefix = "urllib3_future."
        alias_root = "niquests.packages.urllib3"
        real_root = "urllib3_future"
    else:
        alias_prefix = f"niquests.packages.{package}."
        real_prefix = f"{package}."
        alias_root = f"niquests.packages.{package}"
        real_root = package

    _ALIAS_TO_REAL[alias_prefix] = real_prefix
    _ALIAS_TO_REAL[alias_root] = real_root

    # This traversal is apparently necessary such that the identities are
    # preserved (requests.packages.urllib3.* is urllib3.*)
    for mod in list(sys.modules):
        if mod == to_be_imported or mod.startswith(f"{to_be_imported}."):
            inner_mod = mod

            if HAS_LEGACY_URLLIB3 and inner_mod == "urllib3_future" or inner_mod.startswith("urllib3_future."):
                inner_mod = inner_mod.replace("urllib3_future", "urllib3")
            elif inner_mod == "charset_normalizer":
                inner_mod = "chardet"

            try:
                sys.modules[f"niquests.packages.{inner_mod}"] = sys.modules[mod]
            except KeyError:
                continue


class _NiquestsPackagesAliasImporter(importlib.abc.MetaPathFinder):
    """Made to avoid duplicate due to lazy imports at urllib3-future side(...)"""

    def find_spec(
        self,
        fullname: str,
        path: typing.Any = None,
        target: typing.Any = None,
    ) -> importlib.machinery.ModuleSpec | None:
        if fullname in sys.modules:
            return None

        real_name: str | None = None
        for alias, real in _ALIAS_TO_REAL.items():
            if fullname == alias or fullname.startswith(alias if alias.endswith(".") else alias + "."):
                real_name = real + fullname[len(alias) :]
                break

        if real_name is None:
            return None

        # Import the real module first, then point the alias at it.
        real_module = importlib.import_module(real_name)
        sys.modules[fullname] = real_module

        # Return a spec that resolves to the cached module.
        return importlib.util.spec_from_loader(fullname, loader=None, origin=real_name)  # type: ignore[attr-defined]


# Insert at front so we intercept before the default PathFinder.
sys.meta_path.insert(0, _NiquestsPackagesAliasImporter())


__all__ = (
    "urllib3",
    "chardet",
    "charset_normalizer",
    "idna",
)



from __future__ import annotations

import asyncio
import contextlib
import typing


class ASGIWebSocketExtension:
    """Async WebSocket extension for ASGI applications.

    Uses the ASGI websocket protocol with send/receive queues to communicate
    with the application task.
    """

    def __init__(self) -> None:
        self._closed = False
        self._app_send_queue: asyncio.Queue[dict[str, typing.Any]] = asyncio.Queue()
        self._app_receive_queue: asyncio.Queue[dict[str, typing.Any]] = asyncio.Queue()
        self._task: asyncio.Task[None] | None = None

    async def start(self, app: typing.Any, scope: dict[str, typing.Any]) -> None:
        """Start the ASGI app task and perform the WebSocket handshake."""

        async def receive() -> dict[str, typing.Any]:
            return await self._app_receive_queue.get()

        async def send(message: dict[str, typing.Any]) -> None:
            await self._app_send_queue.put(message)

        self._task = asyncio.create_task(app(scope, receive, send))

        # Send connect and wait for accept
        await self._app_receive_queue.put({"type": "websocket.connect"})

        message = await self._app_send_queue.get()

        if message["type"] == "websocket.close":
            self._closed = True
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task
            raise ConnectionError(f"WebSocket connection rejected with code {message.get('code', 1000)}")

        if message["type"] != "websocket.accept":
            self._closed = True
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task
            raise ConnectionError(f"Unexpected ASGI message during handshake: {message['type']}")

    @property
    def closed(self) -> bool:
        return self._closed

    async def next_payload(self) -> str | bytes | None:
        """Await the next message from the ASGI WebSocket app.
        Returns None when the app closes the connection."""
        if self._closed:
            raise OSError("The WebSocket extension is closed")

        message = await self._app_send_queue.get()

        if message["type"] == "websocket.send":
            if "text" in message:
                return message["text"]
            if "bytes" in message:
                return message["bytes"]
            return b""

        if message["type"] == "websocket.close":
            self._closed = True
            return None

        return None

    async def send_payload(self, buf: str | bytes) -> None:
        """Send a message to the ASGI WebSocket app."""
        if self._closed:
            raise OSError("The WebSocket extension is closed")

        if isinstance(buf, (bytes, bytearray)):
            await self._app_receive_queue.put({"type": "websocket.receive", "bytes": bytes(buf)})
        else:
            await self._app_receive_queue.put({"type": "websocket.receive", "text": buf})

    async def close(self) -> None:
        """Close the WebSocket and clean up the app task."""
        if self._closed:
            return

        self._closed = True

        await self._app_receive_queue.put({"type": "websocket.disconnect", "code": 1000})

        if self._task is not None and not self._task.done():
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._task



from __future__ import annotations

from .typing import RetryType, TimeoutType

#: Default timeout (total) assigned for GET, HEAD, and OPTIONS methods.
READ_DEFAULT_TIMEOUT: TimeoutType = 30
#: Default timeout (total) assigned for DELETE, PUT, PATCH, and POST.
WRITE_DEFAULT_TIMEOUT: TimeoutType = 120

DEFAULT_POOLBLOCK: bool = False
DEFAULT_POOLSIZE: int = 10
DEFAULT_RETRIES: RetryType = 0


# we don't want to eagerly load this as some user just
# don't leverage ssl anyway. this should make niquests
# import generally faster.
def __getattr__(name: str):
    if name == "DEFAULT_CA_BUNDLE":
        import wassima

        val = wassima.generate_ca_bundle()
        globals()["DEFAULT_CA_BUNDLE"] = val
        return val
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



"""Tests for rate limiter hooks."""

from __future__ import annotations

import asyncio
import concurrent.futures
import time
from functools import partial

import pytest

import niquests
from niquests.hooks import (
    AsyncLeakyBucketLimiter,
    AsyncTokenBucketLimiter,
    LeakyBucketLimiter,
    TokenBucketLimiter,
)
from tests.testserver.server import Server


class TestSyncLimiters:
    """Tests for sync rate limiters."""

    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(LeakyBucketLimiter, rate=10.0), id="leaky_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=10.0), id="token_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=10.0, capacity=20.0), id="token_bucket_with_capacity"),
        ],
    )
    def test_basic_request(self, limiter_factory):
        """Rate limiter should not prevent basic requests."""
        limiter = limiter_factory()
        with Server.basic_response_server() as (host, port):
            with niquests.Session(hooks=limiter) as session:
                response = session.get(f"http://{host}:{port}")
                assert response.status_code == 200

    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(LeakyBucketLimiter, rate=10.0), id="leaky_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=10.0, capacity=10.0), id="token_bucket"),
        ],
    )
    def test_multiple_requests(self, limiter_factory):
        """Multiple requests should succeed with rate limiting."""
        limiter = limiter_factory()
        with Server.basic_response_server(requests_to_handle=3) as (host, port):
            with niquests.Session(hooks=limiter, headers={"Connection": "close"}) as session:
                for _ in range(3):
                    response = session.get(f"http://{host}:{port}")
                    assert response.status_code == 200

    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(LeakyBucketLimiter, rate=20.0), id="leaky_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=20.0, capacity=20.0), id="token_bucket"),
        ],
    )
    def test_thread_safety(self, limiter_factory):
        """Rate limiter should be thread-safe."""
        limiter = limiter_factory()
        results = []

        def make_request(session, url):
            response = session.get(url)
            return response.status_code

        with Server.basic_response_server(requests_to_handle=10) as (host, port):
            url = f"http://{host}:{port}"
            with niquests.Session(hooks=limiter, headers={"Connection": "close"}) as session:
                with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                    futures = [executor.submit(partial(make_request, session, url)) for _ in range(10)]
                    for future in concurrent.futures.as_completed(futures):
                        results.append(future.result())

        assert all(status == 200 for status in results)
        assert len(results) == 10


class TestSyncLimitersUnit:
    """Unit tests for sync rate limiters without network calls."""

    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(LeakyBucketLimiter, rate=100.0), id="leaky_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=100.0, capacity=20.0), id="token_bucket"),
        ],
    )
    def test_concurrent_calls_are_serialized(self, limiter_factory):
        """Concurrent calls should be properly serialized by the lock."""
        limiter = limiter_factory()
        call_count = 0

        def make_call():
            nonlocal call_count
            limiter.pre_request(None)
            call_count += 1

        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            futures = [executor.submit(make_call) for _ in range(10)]
            concurrent.futures.wait(futures)

        assert call_count == 10

    @pytest.mark.parametrize(
        "limiter_factory,expected_delay",
        [
            pytest.param(partial(LeakyBucketLimiter, rate=10.0), 0.1, id="leaky_bucket"),
            pytest.param(partial(TokenBucketLimiter, rate=10.0, capacity=1.0), 0.1, id="token_bucket"),
        ],
    )
    def test_rate_limiting_introduces_delay(self, limiter_factory, expected_delay):
        """Rate limiting should introduce delays when limit is exceeded."""
        limiter = limiter_factory()

        # First request - immediate
        limiter.pre_request(None)

        # Second request should be delayed
        start = time.monotonic()
        limiter.pre_request(None)
        elapsed = time.monotonic() - start

        # Should have waited approximately expected_delay (with tolerance)
        assert elapsed >= expected_delay * 0.7

    def test_token_bucket_burst_no_wait(self):
        """Token bucket burst requests within capacity should not wait."""
        limiter = TokenBucketLimiter(rate=1.0, capacity=10.0)

        start = time.monotonic()
        for _ in range(5):
            limiter.pre_request(None)
        elapsed = time.monotonic() - start

        # All 5 should complete almost instantly since we have 10 tokens
        assert elapsed < 0.1

    def test_token_bucket_tokens_capped_at_capacity(self):
        """Token bucket tokens should not exceed capacity."""
        limiter = TokenBucketLimiter(rate=100.0, capacity=5.0)

        time.sleep(0.1)
        with limiter._lock:
            limiter._acquire_token()

        assert limiter.tokens <= limiter.capacity


class TestAsyncLimiters:
    """Tests for async rate limiters."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(AsyncLeakyBucketLimiter, rate=10.0), id="leaky_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=10.0), id="token_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=10.0, capacity=20.0), id="token_bucket_with_capacity"),
        ],
    )
    async def test_basic_request(self, limiter_factory):
        """Rate limiter should not prevent basic requests."""
        limiter = limiter_factory()
        with Server.basic_response_server() as (host, port):
            async with niquests.AsyncSession(hooks=limiter) as session:
                response = await session.get(f"http://{host}:{port}")
                assert response.status_code == 200

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(AsyncLeakyBucketLimiter, rate=10.0), id="leaky_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=10.0, capacity=10.0), id="token_bucket"),
        ],
    )
    async def test_multiple_requests(self, limiter_factory):
        """Multiple requests should succeed with rate limiting."""
        limiter = limiter_factory()
        with Server.basic_response_server(requests_to_handle=3) as (host, port):
            async with niquests.AsyncSession(hooks=limiter, headers={"Connection": "close"}) as session:
                for _ in range(3):
                    response = await session.get(f"http://{host}:{port}")
                    assert response.status_code == 200

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(AsyncLeakyBucketLimiter, rate=20.0), id="leaky_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=20.0, capacity=20.0), id="token_bucket"),
        ],
    )
    async def test_concurrent_requests(self, limiter_factory):
        """Rate limiter should handle concurrent async requests."""
        limiter = limiter_factory()
        with Server.basic_response_server(requests_to_handle=10) as (host, port):
            url = f"http://{host}:{port}"
            async with niquests.AsyncSession(hooks=limiter, headers={"Connection": "close"}) as session:

                async def make_request():
                    response = await session.get(url)
                    return response.status_code

                tasks = [make_request() for _ in range(10)]
                results = await asyncio.gather(*tasks)

            assert all(status == 200 for status in results)
            assert len(results) == 10


class TestAsyncLimitersUnit:
    """Unit tests for async rate limiters without network calls."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limiter_factory",
        [
            pytest.param(partial(AsyncLeakyBucketLimiter, rate=100.0), id="leaky_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=100.0, capacity=20.0), id="token_bucket"),
        ],
    )
    async def test_concurrent_calls_are_serialized(self, limiter_factory):
        """Concurrent calls should be properly serialized by the lock."""
        limiter = limiter_factory()
        call_count = 0

        async def make_call():
            nonlocal call_count
            await limiter.pre_request(None)
            call_count += 1

        await asyncio.gather(*[make_call() for _ in range(10)])
        assert call_count == 10

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "limiter_factory,expected_delay",
        [
            pytest.param(partial(AsyncLeakyBucketLimiter, rate=10.0), 0.1, id="leaky_bucket"),
            pytest.param(partial(AsyncTokenBucketLimiter, rate=10.0, capacity=1.0), 0.1, id="token_bucket"),
        ],
    )
    async def test_rate_limiting_introduces_delay(self, limiter_factory, expected_delay):
        """Rate limiting should introduce delays when limit is exceeded."""
        limiter = limiter_factory()

        # First request - immediate
        await limiter.pre_request(None)

        # Second request should be delayed
        start = time.monotonic()
        await limiter.pre_request(None)
        elapsed = time.monotonic() - start

        # Should have waited approximately expected_delay (with tolerance)
        assert elapsed >= expected_delay * 0.7

    @pytest.mark.asyncio
    async def test_token_bucket_burst_no_wait(self):
        """Token bucket burst requests within capacity should not wait."""
        limiter = AsyncTokenBucketLimiter(rate=1.0, capacity=10.0)

        start = time.monotonic()
        for _ in range(5):
            await limiter.pre_request(None)
        elapsed = time.monotonic() - start

        # All 5 should complete almost instantly since we have 10 tokens
        assert elapsed < 0.1

    @pytest.mark.asyncio
    async def test_token_bucket_tokens_capped_at_capacity(self):
        """Token bucket tokens should not exceed capacity."""
        limiter = AsyncTokenBucketLimiter(rate=100.0, capacity=5.0)

        await asyncio.sleep(0.1)
        async with limiter._lock:
            limiter._acquire_token()

        assert limiter.tokens <= limiter.capacity
